/*
Copyright (C) 2011 Mark Chandler (Desura Net Pty Ltd)
Copyright (C) 2014 Bad Juju Games, Inc.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.

Contact us at legal@badjuju.com.

*/

#include "Common.h"
#include "LoginForm.h"

#include "MainApp.h"
#include "PasswordReminder.h"
#include "StripMenuButton.h"
#include "NewAccountDialog.h"
#include "AltLoginDialog.h"

#include "managers/CVar.h"
#include <wx/msgdlg.h>

#include <branding/branding.h>

bool validateUsernameChange(const CVar*, const char*);

CVar gc_savelogin("gc_savelogin", "0");
#ifdef WIN32
CVar gc_saveProxyOff("gc_saveproxyoff", "0");
#endif
CVar gc_saveBypassSSLRevocationCheck( "gc_saveBypassSSLRevocationCheck", "0" );
CVar gc_saveusername("gc_saveusername", "1");

CVar gc_lastusername("gc_lastusername", "", CFLAG_NOFLAGS, &validateUsernameChange);
CVar gc_lastavatar("gc_lastavatar", "");

CVar gc_login_x("gc_login_x","-1", CFLAG_WINUSER);
CVar gc_login_y("gc_login_y","-1", CFLAG_WINUSER);

CVar gc_login_stage_url("gc_login_stage_url", "", CFLAG_WINUSER);
CVar gc_login_stage_last("gc_login_stage_last", "Prod", CFLAG_WINUSER);

bool validateUsernameChange(const CVar* cvar, const char* newValue)
{
	if (gcString(newValue) == gcString(cvar->getString()))
		return false;

	gc_lastavatar.setValue("");
	return true;
}


#define GetHrgnOf(rgn)          ((HRGN)(rgn).GetHRGN())

///////////////////////////////////////////////////////////////////////////////
/// Login Thread
///////////////////////////////////////////////////////////////////////////////

class LoginThread : public Thread::BaseThread
{
public:
	LoginThread(const char* username, const char* password, bool bAltProvider, LoginForm* caller)
		: BaseThread( "Login Thread")
		, m_szUsername(username)
		, m_szPassword(password)
		, m_pCaller(caller)
		, m_bAltProvider(bAltProvider)
	{
	}

	~LoginThread()
	{
		stop();
	}

protected:
	void run()
	{
		try
		{
			MainAppProviderI *pProvider = dynamic_cast<MainAppProviderI*>(g_pMainApp);

			if (m_bAltProvider)
			{
				gcString strProv = gc_login_stage_url.getString();
				pProvider->setProvider(strProv.c_str());
			}
			else
			{
				pProvider->setProvider(nullptr);
			}

			g_pMainApp->logIn(m_szUsername.c_str(), m_szPassword.c_str());

			if (m_pCaller)
				m_pCaller->onLoginEvent();
		}
		catch (gcException &except)
		{
			if (m_pCaller)
				m_pCaller->onLoginErrorEvent(except);
		}
	}

	gcString m_szUsername;
	gcString m_szPassword;
	bool m_bAltProvider;

private:
	LoginForm* m_pCaller;
};



class LoginLink : public StripMenuButton
{
public:
	LoginLink(wxWindow* parent, wxString text, uint32 style) : StripMenuButton(parent, text)
	{
		wxColor linkColor = (wxColour)GetThemeManager().getColor("formlogin", "link-fg");
		wxColor hoverColor = (wxColour)GetThemeManager().getColor("formlogin", "link-hov-fg");

		Managers::LoadTheme(this, "formlogin");

		SetWindowStyle(style);

		setColors(linkColor, hoverColor);
		showFocusBox(hoverColor);
		init(nullptr);

		Bind(wxEVT_CHAR, &LoginLink::onChar, this);
	}

protected:
	void onChar(wxKeyEvent &e)
	{
		if (e.m_keyCode == WXK_RETURN)
			sendClickCommand();
		else
			e.Skip();
	}
};
#ifdef NIX
static CVar gc_allow_wm_positioning("gc_allow_wm_positioning", "true");
#endif


///////////////////////////////////////////////////////////////////////////////
/// Class LoginForm
///////////////////////////////////////////////////////////////////////////////


LoginForm::LoginForm(wxWindow* parent)
	: gcFrame(parent, wxID_ANY, Managers::GetString(L"#LF_TITLE"), wxDefaultPosition, wxSize(420,286), wxCAPTION|wxCLOSE_BOX|wxSYSTEM_MENU|wxWANTS_CHARS|wxMINIMIZE_BOX, true)
{
	m_comboProvider = nullptr;

	m_bAutoLogin = false;
	m_pNewAccount = nullptr;

	Bind(wxEVT_COMMAND_TEXT_ENTER, &LoginForm::onTextBoxEnter, this);
	Bind(wxEVT_CLOSE_WINDOW, &LoginForm::onClose, this);
	Bind(wxEVT_COMMAND_BUTTON_CLICKED, &LoginForm::onButtonClick, this);
	Bind(wxEVT_MOVE, &LoginForm::onMove, this);


	m_imgLogo = new gcImageControl(this, wxID_ANY, wxDefaultPosition, wxSize( 253,81 ));
	m_imgAvatar = new gcImageControl(this, wxID_ANY, wxDefaultPosition, wxSize( 66,66 ));

	m_imgLogo->setImage("#login_logo");
	m_imgAvatar->setImage("#login_default_avatar");

	m_tbUsername = new gcTextCtrl(this, wxID_ANY, Managers::GetString(L"#LF_USER"), wxDefaultPosition, wxDefaultSize, wxTE_PROCESS_ENTER|wxWANTS_CHARS);
	m_tbPassword = new gcTextCtrl(this, wxID_ANY, "", wxDefaultPosition, wxDefaultSize, wxTE_PROCESS_ENTER|wxWANTS_CHARS|wxTE_PASSWORD);
	m_tbPasswordDisp = new gcTextCtrl(this, wxID_ANY, Managers::GetString(L"#LF_PASS"));

	gcString strStagingUrl(gc_login_stage_url.getString());
	gcString strStagingLast(gc_login_stage_last.getString());

	if (strStagingUrl.size() > 0)
	{
		wxArrayString choices;
		choices.Add("Prod");
		choices.Add("Test");

		m_comboProvider = new gcChoice(this, wxID_ANY, wxDefaultPosition, wxDefaultSize, choices);

		if (strStagingLast == "Test")
			m_comboProvider->SetSelection(1);
		else
			m_comboProvider->SetSelection(0);
	}

	m_cbRemPass = new gcCheckBox(this, wxID_ANY, Managers::GetString(L"#LF_AUTO"));
	m_cbRemPass->SetToolTip(Managers::GetString(L"#LF_AUTO_TOOLTIP"));

#ifdef WIN32
	m_cbProxyOff = new gcCheckBox(this, wxID_ANY, Managers::GetString(L"#LF_PROXYOFF"));
	m_cbProxyOff->SetToolTip(Managers::GetString(L"#LF_PROXYOFF_TOOLTIP"));
#endif

	m_cbBypassSSLRevocationCheck = new gcCheckBox( this, wxID_ANY, Managers::GetString( L"#LF_REVOCATION" ) );
	m_cbBypassSSLRevocationCheck->SetToolTip( Managers::GetString( L"#LF_REVOCATION_TOOLTIP" ) );

	m_butSignin = new gcButton(this, wxID_ANY, Managers::GetString(L"#LOGIN"), wxDefaultPosition, wxDefaultSize, wxWANTS_CHARS);
	m_butCancel = new gcButton(this, wxID_ANY, Managers::GetString(L"#CANCEL"), wxDefaultPosition, wxDefaultSize, wxWANTS_CHARS);

	m_linkOffline = new LoginLink(this, Managers::GetString(L"#LF_OFFLINE"), wxALIGN_LEFT);
	m_linkNewAccount = new LoginLink(this, Managers::GetString(L"#LF_NEWACCOUNT"), wxALIGN_CENTER);
	m_linkLostPassword = new LoginLink( this, Managers::GetString(L"#LF_LOSTPASS"), wxALIGN_RIGHT);


	m_butTwitter = new gcImageButton(this, wxID_ANY, wxDefaultPosition, wxSize(42, 42));
	m_butSteam = new gcImageButton(this, wxID_ANY, wxDefaultPosition, wxSize(42, 42));
	m_butFacebook = new gcImageButton(this, wxID_ANY, wxDefaultPosition, wxSize(42, 42));
	m_butGoogle = new gcImageButton(this, wxID_ANY, wxDefaultPosition, wxSize(42, 42));

	m_butTwitter->setDefaultImage("#login_twitter");
	m_butSteam->setDefaultImage("#login_steam");
	m_butFacebook->setDefaultImage("#login_facebook");
	m_butGoogle->setDefaultImage("#login_google");

	m_butTwitter->SetCursor(wxCursor(wxCURSOR_HAND));
	m_butSteam->SetCursor(wxCursor(wxCURSOR_HAND));
	m_butFacebook->SetCursor(wxCursor(wxCURSOR_HAND));
	m_butGoogle->SetCursor(wxCursor(wxCURSOR_HAND));

	m_butTwitter->Bind(wxEVT_COMMAND_BUTTON_CLICKED, &LoginForm::onAltLoginClick, this);
	m_butSteam->Bind(wxEVT_COMMAND_BUTTON_CLICKED, &LoginForm::onAltLoginClick, this);
	m_butFacebook->Bind(wxEVT_COMMAND_BUTTON_CLICKED, &LoginForm::onAltLoginClick, this);
	m_butGoogle->Bind(wxEVT_COMMAND_BUTTON_CLICKED, &LoginForm::onAltLoginClick, this);


	//focus gets set to the first child. Redirect input to username box
	m_imgLogo->Bind(wxEVT_CHAR, &LoginForm::onFormChar, this);
	m_imgAvatar->Bind(wxEVT_CHAR, &LoginForm::onFormChar, this);

#ifdef WIN32
	Bind(wxEVT_LEFT_UP, &LoginForm::onMouseUp, this);
	Bind(wxEVT_LEFT_DOWN, &LoginForm::onMouseDown, this);
	Bind(wxEVT_MOTION, &LoginForm::onMouseMove, this);

	m_imgLogo->Bind(wxEVT_LEFT_UP, &LoginForm::onMouseUp, this);
	m_imgLogo->Bind(wxEVT_LEFT_DOWN, &LoginForm::onMouseDown, this);
	m_imgLogo->Bind(wxEVT_MOTION, &LoginForm::onMouseMove, this);

	m_imgAvatar->Bind(wxEVT_LEFT_UP, &LoginForm::onMouseUp, this);
	m_imgAvatar->Bind(wxEVT_LEFT_DOWN, &LoginForm::onMouseDown, this);
	m_imgAvatar->Bind(wxEVT_MOTION, &LoginForm::onMouseMove, this);
#endif

	m_tbUsername->Bind(wxEVT_COMMAND_TEXT_UPDATED, &LoginForm::onTextChange, this);
	m_tbUsername->Bind(wxEVT_CHAR, &LoginForm::onChar, this);
	m_tbUsername->Bind(wxEVT_KILL_FOCUS, &LoginForm::onBlur, this);
	m_tbUsername->Bind(wxEVT_SET_FOCUS, &LoginForm::onFocus, this);

	m_tbPassword->Bind(wxEVT_CHAR, &LoginForm::onChar, this);
	m_tbPassword->Bind(wxEVT_KILL_FOCUS, &LoginForm::onBlur, this);

	m_tbPasswordDisp->Bind(wxEVT_COMMAND_TEXT_UPDATED, &LoginForm::onTextChange, this);
	m_tbPasswordDisp->Bind(wxEVT_SET_FOCUS, &LoginForm::onFocus, this);

	m_cbRemPass->Bind(wxEVT_CHAR, &LoginForm::onChar, this);
#ifdef WIN32
	m_cbProxyOff->Bind( wxEVT_COMMAND_CHECKBOX_CLICKED, &LoginForm::onCheckBoxClick, this );
#endif

	m_cbBypassSSLRevocationCheck->Bind( wxEVT_COMMAND_CHECKBOX_CLICKED, &LoginForm::onCheckBoxClick, this );

	m_linkOffline->Bind(wxEVT_CHAR, &LoginForm::onChar, this);
	m_linkNewAccount->Bind(wxEVT_CHAR, &LoginForm::onChar, this);
	m_linkLostPassword->Bind(wxEVT_CHAR, &LoginForm::onChar, this);

	m_linkOffline->Bind(wxEVT_COMMAND_BUTTON_CLICKED, &LoginForm::onLinkClick, this);
	m_linkNewAccount->Bind(wxEVT_COMMAND_BUTTON_CLICKED, &LoginForm::onLinkClick, this);
	m_linkLostPassword->Bind(wxEVT_COMMAND_BUTTON_CLICKED, &LoginForm::onLinkClick, this);


	m_butCancel->Bind(wxEVT_CHAR, &LoginForm::onChar, this);
	m_butSignin->Bind(wxEVT_CHAR, &LoginForm::onChar, this);

	m_tbUsername->SetValue(Managers::GetString(L"#LF_USER"));
	m_tbPasswordDisp->SetValue(Managers::GetString(L"#LF_PASS"));

	m_tbPassword->Show(false);
	m_tbPasswordDisp->Show(true);


#ifdef WIN32
	wxFlexGridSizer* fgSizer4 = new wxFlexGridSizer( 1, 3, 0, 0 );
#else
	wxFlexGridSizer* fgSizer4 = new wxFlexGridSizer( 1, 2, 0, 0 );
#endif
	fgSizer4->AddGrowableCol( 0 );
	fgSizer4->SetFlexibleDirection( wxBOTH );
	fgSizer4->SetNonFlexibleGrowMode( wxFLEX_GROWMODE_SPECIFIED );

	fgSizer4->Add( m_cbRemPass, 0, wxALL|wxALIGN_CENTER_VERTICAL, 5 );

#ifdef WIN32
	fgSizer4->Add( m_cbProxyOff, 0, wxALL | wxALIGN_CENTER_VERTICAL, 5 );
#endif

	fgSizer4->Add( m_cbBypassSSLRevocationCheck, 0, wxALL | wxALIGN_CENTER_VERTICAL, 5 );



	wxFlexGridSizer* fgSizer4a = new wxFlexGridSizer( 1, 2, 0, 0 );
	fgSizer4a->AddGrowableCol( 0 );
	fgSizer4a->SetFlexibleDirection( wxBOTH );
	fgSizer4a->SetNonFlexibleGrowMode( wxFLEX_GROWMODE_SPECIFIED );
	fgSizer4a->Add( m_butSignin, 0, wxTOP | wxBOTTOM | wxLEFT, 5 );
	fgSizer4a->Add( m_butCancel, 0, wxTOP | wxBOTTOM | wxLEFT, 5 );


	wxFlexGridSizer* fgSizer5 = new wxFlexGridSizer( 1, 3, 0, 0 );
	fgSizer5->AddGrowableCol( 0 );
	fgSizer5->AddGrowableCol( 1 );
	fgSizer5->AddGrowableCol( 2 );
	fgSizer5->AddGrowableRow( 0 );
	fgSizer5->SetFlexibleDirection( wxBOTH );
	fgSizer5->SetNonFlexibleGrowMode( wxFLEX_GROWMODE_SPECIFIED );
	fgSizer5->Add( m_linkOffline, 0, wxTOP|wxBOTTOM|wxRIGHT, 5 );
	fgSizer5->Add( m_linkNewAccount, 0, wxALL|wxALIGN_CENTER_HORIZONTAL, 5 );
	fgSizer5->Add( m_linkLostPassword, 0, wxALIGN_RIGHT|wxTOP|wxBOTTOM|wxLEFT, 5 );


	wxFlexGridSizer* fgSizer6 = new wxFlexGridSizer( 5, 1, 0, 0 );
	fgSizer6->AddGrowableCol( 0 );
	fgSizer6->SetFlexibleDirection( wxBOTH );
	fgSizer6->SetNonFlexibleGrowMode( wxFLEX_GROWMODE_SPECIFIED );

	if (m_comboProvider)
	{
		wxFlexGridSizer* fgSizer10 = new wxFlexGridSizer(1, 2, 0, 0);
		fgSizer10->AddGrowableCol(0);
		fgSizer10->Add(m_tbUsername, 0, wxEXPAND, 5);
		fgSizer10->Add(m_comboProvider, 0, wxLEFT, 5);

		fgSizer6->Add(fgSizer10, 0, wxEXPAND | wxLEFT, 5);
	}
	else
	{
		fgSizer6->Add(m_tbUsername, 0, wxEXPAND | wxLEFT, 5);
	}

	fgSizer6->Add( m_tbPassword, 0, wxEXPAND|wxTOP|wxLEFT, 5 );
	fgSizer6->Add( m_tbPasswordDisp, 0, wxEXPAND|wxTOP|wxLEFT, 5 );
	fgSizer6->Add( fgSizer4, 1, wxEXPAND, 5 );
	fgSizer6->Add( fgSizer4a, 1, wxEXPAND, 5 );

	wxFlexGridSizer* fgSizer3 = new wxFlexGridSizer( 1, 2, 0, 0 );
	fgSizer3->AddGrowableCol( 1 );
	fgSizer3->SetFlexibleDirection( wxBOTH );
	fgSizer3->SetNonFlexibleGrowMode( wxFLEX_GROWMODE_SPECIFIED );
	fgSizer3->Add( m_imgAvatar, 0, 0, 5 );
	fgSizer3->Add( fgSizer6, 1, wxALIGN_CENTER_VERTICAL|wxALIGN_CENTER_HORIZONTAL|wxEXPAND, 5 );

	wxFlexGridSizer* fgSizerAltLogin = new wxFlexGridSizer(1, 9, 0, 0);
	fgSizerAltLogin->AddGrowableCol(2);
	fgSizerAltLogin->AddGrowableCol(4);
	fgSizerAltLogin->AddGrowableCol(6);
	fgSizerAltLogin->AddGrowableRow(0);
	fgSizerAltLogin->Add(15, 0, 1, wxEXPAND, 5);
	fgSizerAltLogin->Add(m_butTwitter, 0, 0, 5);
	fgSizerAltLogin->Add(0, 0, 1, wxEXPAND, 5);
	fgSizerAltLogin->Add(m_butSteam, 0, 0, 5);
	fgSizerAltLogin->Add(0, 0, 1, wxEXPAND, 5);
	fgSizerAltLogin->Add(m_butFacebook, 0, 0, 5);
	fgSizerAltLogin->Add(0, 0, 1, wxEXPAND, 5);
	fgSizerAltLogin->Add(m_butGoogle, 0, 0, 5);
	fgSizerAltLogin->Add(15, 0, 1, wxEXPAND, 5);

	wxFlexGridSizer* fgSizer2 = new wxFlexGridSizer( 6, 1, 0, 0 );
	fgSizer2->SetFlexibleDirection( wxBOTH );
	fgSizer2->SetNonFlexibleGrowMode( wxFLEX_GROWMODE_SPECIFIED );
	fgSizer2->Add( m_imgLogo, 1, wxTOP|wxBOTTOM|wxALIGN_CENTER_HORIZONTAL, 5 );
	fgSizer2->Add( fgSizer3, 1, wxEXPAND, 5 );
	fgSizer2->Add(0, 9, 1, wxEXPAND, 5);
	fgSizer2->Add(fgSizer5, 1, wxEXPAND, 5);
	fgSizer2->Add(0, 9, 1, wxEXPAND, 5);
	fgSizer2->Add(fgSizerAltLogin, 1, wxEXPAND, 5);



	wxFlexGridSizer* fgSizer1 = new wxFlexGridSizer( 1, 3, 0, 0 );
	fgSizer1->AddGrowableCol( 0 );
	fgSizer1->AddGrowableCol( 2 );
	fgSizer1->AddGrowableRow( 0 );
	fgSizer1->SetFlexibleDirection( wxBOTH );
	fgSizer1->SetNonFlexibleGrowMode( wxFLEX_GROWMODE_SPECIFIED );
	fgSizer1->Add( 0, 0, 1, wxEXPAND, 5 );
	fgSizer1->Add( fgSizer2, 1, wxEXPAND, 5 );
	fgSizer1->Add( 0, 15, 1, wxEXPAND|wxBOTTOM, 15 );

	this->SetSizer( fgSizer1 );
	this->Layout();


	m_bSavePos = false;

	m_vTabOrder.push_back(m_tbUsername);
	m_vTabOrder.push_back(m_tbPasswordDisp);
	m_vTabOrder.push_back(m_tbPassword);
	m_vTabOrder.push_back( m_cbRemPass );
#ifdef WIN32
	m_vTabOrder.push_back( m_cbProxyOff );
#endif
	m_vTabOrder.push_back( m_cbBypassSSLRevocationCheck );
	m_vTabOrder.push_back(m_butSignin);
	m_vTabOrder.push_back(m_butCancel);
	m_vTabOrder.push_back(m_linkOffline);
	m_vTabOrder.push_back(m_linkNewAccount);
	m_vTabOrder.push_back(m_linkLostPassword);

	Centre(wxCENTRE_ON_SCREEN);
	SetFocus();

	if (gc_login_x.getInt() == -1 || gc_login_y.getInt() == -1)
	{
		int x = wxSystemSettings::GetMetric(wxSYS_SCREEN_X);
		int y = wxSystemSettings::GetMetric(wxSYS_SCREEN_Y);

		int w=0;
		int h=0;
		GetSize(&w, &h);

		int newX = (x-w)/2;
		int newY = (y-h)/2;

		if (newX>0 && newY>0)
			SetSize(newX, newY, -1, -1);
	}
	else
	{
#ifdef NIX
		if (! gc_allow_wm_positioning.getBool())
#endif
			setSavedWindowPos(gc_login_x.getInt(), gc_login_y.getInt(), UINT_MAX, UINT_MAX);
	}

	if (gc_savelogin.getBool())
		m_cbRemPass->SetValue(true);

#ifdef WIN32
	bool isProxyOff = gc_saveProxyOff.getBool();
	UTIL::OS::setProxyOff( isProxyOff );
	m_cbProxyOff->SetValue( isProxyOff );
#endif

	bool isBypassSSLRevocationCheck = gc_saveBypassSSLRevocationCheck.getBool();
	UTIL::OS::setBypassSSLRevocationCheck( isBypassSSLRevocationCheck );
	m_cbBypassSSLRevocationCheck->SetValue( isBypassSSLRevocationCheck );

	if (gc_saveusername.getBool())
	{
		const char* str = gc_lastusername.getString();
		const char* avatar = gc_lastavatar.getString();

		if (gcString(str) != "")
		{
			m_tbUsername->SetValue(str);

			m_tbPasswordDisp->Show(false);
			m_tbPassword->Show();
			m_tbPassword->SetFocus();
			Layout();
		}

		if (avatar && UTIL::FS::isValidFile(UTIL::FS::PathWithFile(avatar)))
			m_imgAvatar->setImage(avatar);
	}

	m_pLogThread = nullptr;
	m_bSavePos = true;
	m_bMouseDrag = false;

	onLoginEvent += guiDelegate(this, &LoginForm::onLogin);
	onLoginErrorEvent += guiDelegate(this, &LoginForm::onLoginError);
	onStartLoginEvent += guiDelegate(this, &LoginForm::onStartLogin);

	Managers::LoadTheme(this, "formlogin");

	Managers::LoadTheme(m_cbRemPass, "formlogin");

#ifdef WIN32
	Managers::LoadTheme(m_cbProxyOff, "formlogin");
#endif

	Managers::LoadTheme( m_cbBypassSSLRevocationCheck, "formlogin" );

	SetSize(wxSize(400,336));

	setFrameRegion();

	onNewAccountLoginEvent += guiDelegate(this, &LoginForm::newAccountLoginCB);
	onNewAccountLoginErrorEvent += guiDelegate(this, &LoginForm::newAccountLoginErrorCB);
}

LoginForm::~LoginForm()
{
	onLoginEvent -= guiDelegate(this, &LoginForm::onLogin);
	onLoginErrorEvent -= guiDelegate(this, &LoginForm::onLoginError);
	onStartLoginEvent -= guiDelegate(this, &LoginForm::onStartLogin);

	safe_delete(m_pLogThread);
}

wxRect LoginForm::getWindowsBorderRect() const
{
	// SetWindowRgn expects the region to be in coordinants
	// relative to the window, not the client area.  Figure
	// out the offset, if any.
#ifdef WIN32
	RECT rect;
	DWORD dwStyle =   ::GetWindowLong((HWND)GetHWND(), GWL_STYLE);
	DWORD dwExStyle = ::GetWindowLong((HWND)GetHWND(), GWL_EXSTYLE);
	::GetClientRect((HWND)GetHWND(), &rect);
	::AdjustWindowRectEx(&rect, dwStyle, ::GetMenu((HWND)GetHWND()) != nullptr, dwExStyle);

	int x=rect.left*-1;
	int y=rect.top*-1;

	int w = rect.right+rect.left;
	int h = rect.bottom+rect.left;//+ (y-x)-1;

	wxPoint point = GetPosition();

	return wxRect(wxPoint(x, y), wxSize(w, h));
#else
	return wxRect(wxPoint(0,0), wxSize(800, 600));
#endif // LINUX TODO

}

void LoginForm::setFrameRegion()
{
#ifdef WIN32
	wxRect wbr = getWindowsBorderRect();
	wxPoint br = wbr.GetBottomRight();

	br.x += 2;
	br.y += 2;

	wxRegion region = CreateRoundRectRgn(wbr.x, wbr.y, br.x, br.y, 5, 5);

	// Windows takes ownership of the region, so
	// we'll have to make a copy of the region to give to it.
	DWORD noBytes = ::GetRegionData(GetHrgnOf(region), 0, nullptr);
	RGNDATA *rgnData = (RGNDATA*) new char[noBytes];
	::GetRegionData(GetHrgnOf(region), noBytes, rgnData);
	HRGN hrgn = ::ExtCreateRegion(nullptr, noBytes, rgnData);
	delete[] (char*) rgnData;

	// Now call the shape API with the new region.
	::SetWindowRgn((HWND)GetHWND(), hrgn, FALSE);
#endif // LINUX TODO
}



void LoginForm::onMouseUp(wxMouseEvent& event)
{
#ifdef WIN32
	if (!m_bMouseDrag)
		return;

	m_bMouseDrag = false;

	if (HasCapture())
		ReleaseMouse();
#else
	event.Skip();
#endif
}

void LoginForm::onMouseDown(wxMouseEvent& event)
{
#ifdef WIN32
	wxPoint pos = wxGetMousePosition()-GetScreenPosition();

	m_bMouseDrag = true;
	m_StartDrag = pos;
	m_StartPos = wxGetMousePosition();

	if (!HasCapture())
		CaptureMouse();
#else
	event.Skip();
#endif
}

void LoginForm::onMouseMove(wxMouseEvent& event)
{
#ifdef WIN32
	if (!m_bMouseDrag)
		return;

	wxPoint pos = wxGetMousePosition();
	SetPosition(pos-m_StartDrag);
#else
	event.Skip();
#endif
}

void LoginForm::onBlur(wxFocusEvent& event)
{
	if (event.GetId() == m_tbUsername->GetId())
	{
		if (m_tbUsername->GetValue() == "")
			m_tbUsername->SetValue(Managers::GetString(L"#LF_USER"));
	}
	else if (event.GetId() == m_tbPassword->GetId())
	{
		if (m_tbPassword->GetValue() == "" && m_tbPassword->IsShown())
		{
			m_tbPassword->Show(false);
			m_tbPasswordDisp->Show(true);
			Layout();
			return;
		}
	}

	event.Skip();
}

void LoginForm::onFocus(wxFocusEvent& event)
{
	if (event.GetId() == m_tbUsername->GetId())
	{
		gcString defaultText = Managers::GetString(L"#LF_USER");

		if (m_tbUsername->GetValue() == defaultText)
			m_tbUsername->SetValue("");
	}
	else if (event.GetId() == m_tbPasswordDisp->GetId())
	{
		if (m_tbPasswordDisp->IsShown())
		{
			m_tbPasswordDisp->Show(false);
			m_tbPassword->Show();
			Layout();
			m_tbPassword->SetFocus();
			return;
		}
	}

	event.Skip();
}

void LoginForm::onTextChange(wxCommandEvent& event)
{
	gcWString defaultText;
	gcTextCtrl* textCtrl;

	if (event.GetId() == m_tbUsername->GetId())
	{
		textCtrl = m_tbUsername;
		defaultText = Managers::GetString(L"#LF_USER");

		if (gc_saveusername.getBool())
		{
			const char* str = gc_lastusername.getString();
			const char* avatar = gc_lastavatar.getString();

			if (gcString(str) != gcString(m_tbUsername->GetValue().c_str().AsWChar()))
			{
				m_imgAvatar->setImage("#login_default_avatar");
			}
			else
			{
				if (avatar && UTIL::FS::isValidFile(UTIL::FS::PathWithFile(avatar)))
					m_imgAvatar->setImage(avatar);
			}
		}
	}
	else if (event.GetId() == m_tbPasswordDisp->GetId())
	{
		textCtrl = m_tbPasswordDisp;
		defaultText = Managers::GetString(L"#LF_PASS");
	}
	else
	{
		return;
	}

	const char* colorName = "tb-fg";

	if (textCtrl->GetValue() == defaultText)
		colorName = "tb-default-fg";

	wxColor color = wxColor(GetThemeManager().getColor("formlogin", colorName));

	textCtrl->SetForegroundColour(color);
	textCtrl->Refresh();
}



void LoginForm::autoLogin()
{
	char* user = nullptr;
	char* pass = nullptr;

	try
	{
		UserCoreGetLoginFN* getLogin = (UserCoreGetLoginFN*)UserCore::FactoryBuilderUC(USERCORE_GETLOGIN);

		if (!getLogin)
			throw gcException();

		getLogin(&user, &pass);

		if (user && pass)
		{
			m_tbUsername->SetValue(user);
			m_tbPassword->SetValue(pass);

			m_tbPasswordDisp->Show(false);
			m_tbPassword->Show();
			Layout();

			m_bAutoLogin = true;
			doLogin();
		}
	}
	catch (gcException &except)
	{
		Msg(gcString("Auto Login Failed: {0}\n", except));
	}

	safe_delete(user);
	safe_delete(pass);
}

void LoginForm::onMove(wxMoveEvent  &event)
{
	if (m_bSavePos)
	{
		gc_login_x.setValue(GetPosition().x);
		gc_login_y.setValue(GetPosition().y);
	}

	event.Skip();
}

void LoginForm::onTextBoxEnter( wxCommandEvent& event )
{
	doLogin();
}

void LoginForm::onClose( wxCloseEvent& event )
{
	if (m_pNewAccount)
		m_pNewAccount->EndModal(0);

	if (m_bDisabled)
	{
		event.Skip();
		return;
	}

	Show(false);

	// if we can veto means it wasnt forced by the parent
	if (event.CanVeto() && GetParent())
		GetParent()->Close(true);
}

void LoginForm::onLinkClick( wxCommandEvent& event )
{
	if ( event.GetId() == m_linkNewAccount->GetId() )
	{
		onNewAccount();
	}
	else if ( event.GetId() == m_linkLostPassword->GetId() )
	{
		PasswordReminder wxPassReminderForm( this );

		m_pNewAccount = &wxPassReminderForm;
		wxPassReminderForm.ShowModal();
		m_pNewAccount = nullptr;
	}
	else if ( event.GetId() == m_linkOffline->GetId() )
	{
		if ( GetParent() )
		{
			Show( false );

			MainApp* temp = dynamic_cast<MainApp*>(GetParent());
			if ( temp )
				temp->offlineMode();
		}
	}
}

void LoginForm::onCheckBoxClick( wxCommandEvent& event )
{
#ifdef WIN32
	if ( event.GetId() == m_cbProxyOff->GetId() )
	{
		bool isProxyOff = event.IsChecked();
		m_cbProxyOff->SetValue( isProxyOff );
		gc_saveProxyOff.setValue( isProxyOff );
		UTIL::OS::setProxyOff( isProxyOff );
	}
	else
#endif
		if ( event.GetId() == m_cbBypassSSLRevocationCheck->GetId() )
		{
			bool isBypassSSLRevocationCheck = event.IsChecked();
			m_cbBypassSSLRevocationCheck->SetValue( isBypassSSLRevocationCheck );
			gc_saveBypassSSLRevocationCheck.setValue( isBypassSSLRevocationCheck );
			UTIL::OS::setBypassSSLRevocationCheck( isBypassSSLRevocationCheck );
		}
}

void LoginForm::onButtonClick(wxCommandEvent& event)
{
	if (event.GetId() == m_butCancel->GetId())
	{
		Close();
	}
	else if (event.GetId() == m_butSignin->GetId())
	{
		doLogin();
	}
}

void LoginForm::doLogin()
{
	gcString user((const wchar_t*)m_tbUsername->GetValue().c_str());
	gcString pass((const wchar_t*)m_tbPassword->GetValue().c_str());

	doLogin(user, pass);
}

void LoginForm::doLogin(gcString user, gcString pass)
{
	ASSERT_UITHREAD();
	if (user == "" || pass == "")
	{
		gcWString errMsg(L"{0}\n", Managers::GetString(L"#LF_VALDERROR"));

		if (user == "")
			errMsg += gcWString(L"\t{0}\n", Managers::GetString(L"#LF_ERRBLANKNAME"));

		if (pass == "")
			errMsg += gcWString(L"\t{0}\n", Managers::GetString(L"#LF_ERRPASSWORD"));

		errMsg += gcWString(L"\n{0}", Managers::GetString(L"#LF_ERRPLEASE"));
		gcMessageBox(this, errMsg, Managers::GetString(L"#LF_ERRTITLE"));

		return;
	}

	if (m_bDisabled)
		return;

	if (m_comboProvider)
		m_comboProvider->Disable();

	m_butTwitter->Disable();
	m_butSteam->Disable();
	m_butFacebook->Disable();
	m_butGoogle->Disable();

	m_tbUsername->Disable();
	m_tbPassword->Disable();
	m_butSignin->Disable();

	m_cbRemPass->Disable();
#ifdef WIN32
	m_cbProxyOff->Disable();
#endif

	m_cbBypassSSLRevocationCheck->Disable();
	m_butCancel->Disable();

	m_linkOffline->Disable();
	m_linkNewAccount->Disable();
	m_linkLostPassword->Disable();

	Refresh(false);

	//password will be user id cookie for autologin
	if (m_bAutoLogin)
	{
		auto l = std::make_pair(pass, gcString(""));
		onStartLoginEvent(l);
	}
	else
	{
		auto l = std::make_pair(user, pass);
		onStartLoginEvent(l);
	}
}

void LoginForm::onStartLogin(std::pair<gcString, gcString> &l)
{
	ASSERT_UITHREAD();
	safe_delete(m_pLogThread);

	gcString user = l.first;
	gcString pass = l.second;

	bool bAltProvider = false;

	if (m_comboProvider)
	{
		int nSelection = m_comboProvider->GetSelection();
		bAltProvider = nSelection != 0;
	}

	m_pLogThread = new LoginThread(user.c_str(), pass.c_str(), bAltProvider, this);
	m_pLogThread->start();
}

void LoginForm::onLogin()
{
	ASSERT_UITHREAD();
	safe_delete(m_pLogThread);

	bool remPass = m_cbRemPass->GetValue();
	gc_savelogin.setValue(remPass);

	if (gc_saveusername.getBool())
	{
		wxString dastr = m_tbUsername->GetValue();
		gc_lastusername.setValue(dastr.ToUTF8());
	}

	SaveCVars();

	if (m_comboProvider)
		gc_login_stage_last.setValue(m_comboProvider->GetStringSelection().ToUTF8());

	Show(false);

	MainApp* temp = dynamic_cast<MainApp*>(GetParent());

	if (temp)
		temp->onLoginAccepted(remPass, m_bAutoLogin);
}

void LoginForm::onLoginError(gcException &e)
{
	ASSERT_UITHREAD();
	safe_delete(m_pLogThread);

	if (!m_bAutoLogin)
		gcErrorBox(this, "#LF_ERRTITLE", "#LF_ERROR", e);
	else
		Msg(gcString("Auto login failed: {0}\n", e));

	if (m_comboProvider)
		m_comboProvider->Enable();

	m_butTwitter->Enable();
	m_butSteam->Enable();
	m_butFacebook->Enable();
	m_butGoogle->Enable();

	m_tbPassword->Clear();

	m_tbUsername->Enable();
	m_tbPassword->Enable();
	m_butSignin->Enable();
	m_butCancel->Enable();

	m_cbRemPass->Enable();
#ifdef WIN32
	m_cbProxyOff->Enable();
#endif

	m_cbBypassSSLRevocationCheck->Enable();
	m_linkOffline->Enable();
	m_linkNewAccount->Enable();
	m_linkLostPassword->Enable();

	//bad password
	if (e.getSecErrId() == 103)
	{
		m_tbPassword->SetFocus();
	}
	else
	{
		m_tbUsername->SetFocus();
		m_tbUsername->SelectAll();
	}

	m_bAutoLogin = false;
}

void LoginForm::onChar(wxKeyEvent& event)
{
	if (event.m_keyCode == WXK_TAB)
	{
		processTab(!event.ShiftDown(), event.GetId());
	}
	else
	{
		if (event.GetId() == m_cbRemPass->GetId())
		{
			if (event.m_keyCode == WXK_RETURN)
				doLogin();
			else if (event.m_keyCode == ' ')
				m_cbRemPass->SetValue(!m_cbRemPass->GetValue());
		}

		if (event.GetKeyCode() == WXK_CONTROL_A)
		{
			if (event.GetId() == m_tbPassword->GetId())
				m_tbPassword->SelectAll();
			else if (event.GetId() == m_tbUsername->GetId())
				m_tbUsername->SelectAll();
		}

		event.Skip();
	}
}

void LoginForm::onFormChar(wxKeyEvent& event)
{
	if (event.m_keyCode == WXK_TAB)
	{
		processTab(!event.ShiftDown(), event.GetId());
	}
	else
	{
		int32 keyCode = event.GetKeyCode();
		if (!event.HasModifiers() && ((keyCode >= '0' && keyCode <= '9') || (keyCode >= 'a' && keyCode <= 'z') || (keyCode >= 'A' && keyCode <= 'Z' )))
		{
			m_tbUsername->SetFocus();
			m_tbUsername->SetValue((char)keyCode);
			m_tbUsername->SetInsertionPointEnd();
		}
		else
		{
			event.Skip();
		}
	}
}

void LoginForm::processTab(bool forward, int32 id)
{
	uint32 index = UINT_MAX;

	for (size_t x=0; x<m_vTabOrder.size(); x++)
	{
		if (m_vTabOrder[x]->GetId() == id)
		{
			index = x;
			break;
		}
	};

	if (index != UINT_MAX)
	{
		uint32 start = index;

		do
		{
			if (forward)
			{
				index++;

				if (index == m_vTabOrder.size())
					index=0;
			}
			else
			{
				if (index == 0)
					index=m_vTabOrder.size();

				index--;
			}

			if (index == start)
				return;
		}
		while (m_vTabOrder[index]->IsShown() == false);

		m_vTabOrder[index]->SetFocus();

		wxTextCtrl* text = dynamic_cast<wxTextCtrl*>(m_vTabOrder[index]);

		if (text)
			text->SetInsertionPointEnd();
	}
	else
	{
		m_tbUsername->SetFocus();
	}
}

void LoginForm::onNewAccount()
{
	ASSERT_UITHREAD();

	gcString strApiUrl;

	if (m_comboProvider && m_comboProvider->GetSelection() != 0)
		strApiUrl = gc_login_stage_url.getString();

	NewAccountDialog naf(this, strApiUrl.c_str());

	m_pNewAccount = &naf;
	naf.ShowModal();
	m_pNewAccount = nullptr;
}

void LoginForm::newAccountLogin(const char* username, const char* cookie)
{
	auto info = std::make_pair(gcString(username), gcString(cookie));
	onNewAccountLoginEvent(info);
}

void LoginForm::newAccountLoginCB(std::pair<gcString, gcString> &info)
{
	ASSERT_UITHREAD();

	gcString username = info.first;
	gcString cookie = info.second;

	if (cookie.empty())
		return;

	if (username.size() == 0 || cookie.size() == 0)
	{
		newAccountLoginError("Invalid data returned from auto login.");
		return;
	}

	if (m_pNewAccount)
		m_pNewAccount->EndModal(0);

	m_tbUsername->SetValue(username);
	m_tbPassword->SetValue(cookie);

	m_tbPasswordDisp->Show(false);
	m_tbPassword->Show();
	Layout();

	doLogin(username, cookie);
}

void LoginForm::newAccountLoginError(const char* szErrorMessage)
{
	gcString strErrorMessage(szErrorMessage);

	if (wxThread::IsMain())
		newAccountLoginErrorCB(strErrorMessage);
	else
		onNewAccountLoginErrorEvent(strErrorMessage);
}

void LoginForm::newAccountLoginErrorCB(gcString &szErrorMessage)
{
	ASSERT_UITHREAD();

	if (m_pNewAccount)
		m_pNewAccount->EndModal(0);

	gcException e(ERR_INVALID, szErrorMessage);
	onLoginError(e);
}

void LoginForm::onAltLoginClick(wxCommandEvent& event)
{
	const char* szProvider = nullptr;

	if (event.GetId() == m_butTwitter->GetId())
		szProvider = "twitter";
	else if (event.GetId() == m_butSteam->GetId())
		szProvider = "steam";
	else if (event.GetId() == m_butFacebook->GetId())
		szProvider = "facebook";
	else if (event.GetId() == m_butGoogle->GetId())
		szProvider = "google";

	onAltLogin(szProvider);
}

void LoginForm::onAltLogin(const char* szProvider)
{
	ASSERT_UITHREAD();

	if (szProvider == nullptr)
		return;

	gcString strApiUrl;

	if (m_comboProvider && m_comboProvider->GetSelection() != 0)
		strApiUrl = gc_login_stage_url.getString();

	AltLoginDialog naf(this, szProvider, strApiUrl.c_str());

	m_pNewAccount = &naf;
	naf.ShowModal();
	m_pNewAccount = nullptr;
}

void LoginForm::showForcedLogoutPromt()
{
	gcException e(ERR_NOUSER, "Your login has expired, please re-login.");
	gcErrorBox(this, "#LF_ERRTITLE", "#LF_ERROR", e);
}
